import asyncio
import sys
import json
import time
import urllib
import threading

from pycfhelpers.logger import log
from typing import Optional

try:
    import tornado
    from tornado.web import RequestHandler, Application
except ImportError:
    log.warning("[TORNADO] Tornado not installed, external HTTP server will be unavailable")


class Http:
    """
    Cellframe-compatible request wrapper.
    """

    def __init__(self, trnd: RequestHandler):
        self.trnd = trnd
        self._body_data: Optional[dict] = None
        log.notice(f"[TORNADO] Http wrapper initialized for request: {trnd.request.method} {trnd.request.uri}")

    @property
    def query(self) -> str:
        query = self.trnd.request.query
        log.notice(f"[TORNADO] Raw query string: {query}")
        return query

    @property
    def body(self) -> dict:
        if self._body_data is None:
            try:
                self._body_data = json.loads(self.trnd.request.body.decode())
                log.notice(f"[TORNADO] Parsed JSON body: {self._body_data}")
            except Exception as e:
                log.warning(f"[TORNADO] Failed to parse JSON body: {e}")
                self._body_data = {}
        return self._body_data

    def replyAdd(self, s: str):
        log.notice(f"[TORNADO] Sending response: {s[:500]}...")  # trimmed to 500 chars
        self.trnd.write(s)


class HttpCode:
    def __init__(self):
        self.code = 200
        log.notice("[TORNADO] HttpCode initialized with default 200")

    def set(self, v: int):
        self.code = v
        log.notice(f"[TORNADO] HttpCode set to {v}")

    def get(self) -> int:
        return self.code



def build_main_handler(handler_fn):
    class MainHandler(RequestHandler):
        async def prepare(self):
            self.start_time = time.monotonic()
            self.set_header("Content-Type", "application/json")
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Headers", "Content-Type, Authorization")
            self.set_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")

            log.notice(f"[TORNADO] Received {self.request.method} {self.request.uri}")

        async def get(self):
            await self._handle()

        async def post(self):
            await self._handle()

        async def options(self):
            self.set_status(204)
            self.finish()

        async def _handle(self):
            http = Http(self)
            code = HttpCode()

            try:
                if asyncio.iscoroutinefunction(handler_fn):
                    log.debug("[TORNADO] Detected async handler")
                    await handler_fn(http, code)
                else:
                    log.debug("[TORNADO] Detected sync handler, running in executor")
                    loop = asyncio.get_event_loop()
                    await loop.run_in_executor(None, handler_fn, http, code)
            except Exception as e:
                log.error(f"[TORNADO] Handler error: {e}")
                self.set_status(500)
                self.finish(json.dumps({"error": str(e)}))
                return

            status = code.get()
            log.notice(f"[TORNADO] Final HTTP status: {status}")
            self.set_status(status)
            self.finish()

        def on_finish(self):
            duration = time.monotonic() - self.start_time
            log.notice(f"[TORNADO] {self.request.method} {self.request.uri} -> {self.get_status()} in {duration:.3f}s")

    return MainHandler

    

def init_tornado_http(url: str, handler, port: int):
    """
    Launch Tornado server in separate thread.
    """
    if "tornado" not in sys.modules:
        raise BaseException("[TORNADO] Tornado not installed")

    async def eventloop():
        route_pattern = rf"{url}.*"
        log.notice(f"[TORNADO] Tornado routing pattern: {route_pattern}")
        MainHandler = build_main_handler(handler)
        app = Application([(route_pattern, MainHandler)])
        app.listen(port)
        log.notice(f"[TORNADO] Server is listening on port {port}")
        await asyncio.Event().wait()

    def run():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(eventloop())

    log.notice(f"[TORNADO] Launching HTTP server on port {port}, url: {url}")
    threading.Thread(target=run, daemon=True).start()